#!/usr/bin/perl -w
use diagnostics;
use strict;
use CGI;
use CGI::Carp qw(carp fatalsToBrowser);
use HTML::Entities();

use ezApp::DB;
use ezApp::Interpreter;

package ezApp::job;	

my $Hackable = 1;

#---------------------------------------------------------------
#
# 		Main script for ezApp application framework.
#
# 	Copyright (c) 2001 Richard K. Moore.
#				richard@cyberjournal.org
#
#	This script is free software; you can redistribute it and/or
#	modify it under the same terms as Perl itself.
#
#-------------------------------------------------------------------
# CHANGE LOG
#
#		11/10/01 rkm
#			- set $Hackable=1.
#-------------------------------------------------------------------
#
#		This module implements the top-level control logic for the 
#		application.  Our hash provides storage for shared ezApp 
#		system variables.  
#
#-----------------------------------------------------------------------------------------------------#

my @ezbuttonAttrVarNames = qw(name label screen persistentvars nonpersistentvars tagtext);
my @ReservedVarNames = qw(screenname persistentvars editedvars jobnotes jobwarnings);
my $ReservedVarNames = { map {$_=>1} @ReservedVarNames };

#-----------------------------------------------------------------------------------------------------#
sub new 
{
	my $class = shift;
	my $self = {
		query 								=> CGI->new(),				# CGI $query object
		vars 									=> {},								# hash which mirrors $query-param()
		subscreens						=> {},								# subscreens which have been brought in by <use>
		ezbuttons							=> {},								# information about ezbuttons which have been encountered
		persistentvars 				=> {},								# Keys of this hash are the currently persistent $vars values.
		editedvars 						=> {},								# Keys of this hash are the $vars values which appear in screen fields.
		myparsValuesAttr 			=> {},								# Par values for next screen's <my> vars. Comes from either URL? or last button hit.	
		ParValTemp						=> '',								# This variable is included in the namespace of every subscreen, enabling
																								# Interpreter to interact with the par space of subscreens.
		fhTrace								=> '',								# the file handle to which trace information is written
		ReservedVarNames  		=> $ReservedVarNames,	# $vars keys reserved for use by ezApp kernel.
		NotOkDieMessage				=> 'job.pm initialization failed',
		isopen								=> 1
	};
	bless $self, $class;
$self}
#----------------------------------------------------------------------
sub dieClean
{
	my ($self, $Message) = @_;
	$self->{Died}++;
	$self->{DieMessage} = join "<br>\n", 
		($self->{DieMessage} ? $self->{DieMessage} : () ),  
			( $Message ? $Message : () );
''}
#-----------------------------------------------------------------------------------------------------#
sub Load_vars			# Load $vars from $query and clear $query.
									# Establish first-time default $vars->{screenname, persistentvars}.
{
	my $self 						= shift;
	my $query 					= $self->{query};
	my $vars 						= $self->{vars};
	my $PersistentVars	= $self->{PersistentVars};
	my @FormFieldNames 	= $query->param();
	
	# Move $query->param to $vars.
	# Turn list values into array refs.
	
		foreach my $FormFieldName (@FormFieldNames)
		{
			my @FormFieldValues = $query->param($FormFieldName);
			$vars->{$FormFieldName} = @FormFieldValues > 1 
				? [@FormFieldValues]
				: shift @FormFieldValues;
			$query->param(-name=>$FormFieldName, -value=>'');
		};
		
	if ( $vars->{screenname} )
	{
		# We are resuming an ongoing session.
	
			# Reconstruct PersistentVars hash.
			
				$self->{PersistentVars} = { map { $_ => 1 } split ",", $vars->{persistentvars} };
				delete $vars->{persistentvars};
			
			# Eliminate spurious hacker flags.
			
				unless ($Hackable)
					{ 
						foreach ('URLTraceOn','URLFlushEachTrace', 'URLScreenName', 'URLNoEmail') 
							{ delete $vars->{$_} } 
					};
					
			# Let subclass do its own initialization.
			
				$self->{SessionSetupOverrideScreen} = $self->ContinueSessionSetup;
					
			return;
	};
	
	
	
	## We are initiating a new session.
	
		# Ensure that compiled screens are up to date
		
			$self->CompileScreensIfOutOfDate;
	
		
	
		# Move $vars entries to keys "ezpar:$_"
		# for initialization of intial screen's pars.
		
			foreach (keys %$vars)
			{ 
				$vars->{"ezpar:$_"} = $vars->{$_}; 
				delete $vars->{$_};
				$PersistentVars->{"ezpar:$_"}++
			};
				
		# Setup our own persistent vars.
		
			foreach (@ReservedVarNames) 
				{ $vars->{$_} = '' };
			$vars->{persistentvars} = join ',', @ReservedVarNames;
			
		
		if ($Hackable)
		{
			# Process Hacker flags.
			
				foreach ('TraceOn','FlushEachTrace', 'ScreenName', 'NoEmail')
				{
					$vars->{"URL$_"} = $vars->{"ezpar:$_"}; 
					delete  $vars->{"ezpar:$_"};
					$vars->{persistentvars} = join ',', $vars->{persistentvars}, "URL$_";
				};
		};
		
		# Establish PersistentVars.
	
			$self->{PersistentVars} = { map { $_ => 1 } split ",", $vars->{persistentvars} };
			delete $vars->{persistentvars};
			
		# Let subobject job do its own initializing and overriding.
		
			$vars->{screenname} = $self->StartNewSessionSetup || 'index';
}
#-----------------------------------------------------------------------------------------------------#
sub StartNewSessionSetup {'index'}		# default base method just goes to index.ezscr.
#-----------------------------------------------------------------------------------------------------#
sub ContinueSessionSetup {''}					# default base method does not override screen.
#-----------------------------------------------------------------------------------------------------#
sub DoApplication		# This is the top-level application logic.
										# Our job is to figure out which screen comes next,
										# get it interpreted, and print it to stdout.
{
	my $self 						= shift;
	my $vars 						= $self->{vars};
	my $query 					= $self->{query};
	my $Interpreter 		= $self->{Interpreter}
											= ezApp::Interpreter->new($self);
	
	$self->{isopen}
		|| die $self->{NotOkDieMessage};

	# Load $query->param() into $vars.
	# Session start will be recognized and intialization performed.
	
		$self->Load_vars;
		my $varScanString = "\t".( join "\t", keys %$vars )."\t";
	
	# Setup tracing.
	
		if ($vars->{URLTraceOn})
		{
			use FileHandle;
			my $fhTrace = FileHandle->new;
			(open $fhTrace, '>>ParseTrace.txt' )
				|| die "Could not open >ParseTrace.txt: $!\n";
			$self->{fhTrace} = $fhTrace;
		};

	# Recover and interpret attributes of last ezbutton hit (if any) and get any screen par settings.
	
		my $ParValuesAttr = $self->LoadButtonAttributes($varScanString);
		
		
	## Determine which screen is desired by the app.
		
		my $DesiredScreenName = eval
			{
				# $self->trace("checking URLScreenName");
				my $name = $vars->{URLScreenName};
				delete $vars->{URLScreenName};
				return $name if $name;
				
				# $self->trace("checking SessionSetupOverrideScreen");
				$name = $self->{SessionSetupOverrideScreen};
				return $name if $name;
				
				# $self->trace("checking LastHitButtonScreen");
				$name = $self->{LastHitButtonScreen};
				return $name if $name;
				
				my $LastHitButtonName = $self->{LastHitButtonName};
				my $ButtonMethod = "${LastHitButtonName}_button";
				if ( $self->can($ButtonMethod) )
				{
					# $self->trace("calling ButtonMethod = $ButtonMethod");
					$name = $self->$ButtonMethod();
					# $self->trace("'$ButtonMethod' returned screen '$name'");
					return $name if $name;
				};
				
				# $self->trace("using vars->{screenname} = $vars->{screenname}");
				return $vars->{screenname}
			};
			
		! $@
			|| die "Syntax error in job->DoApplication desired screen:\n$@";
			
		# $self->trace("Desired screen name = $DesiredScreenName");
	
	
	## Perform setup for DesiredScreen, which may redirect us.
		
		my $DisplayableScreenName = eval
			{
				# Give DesiredScreen a chance.
				
					my $DesiredScreenMethod = "${DesiredScreenName}_screen";
					return $DesiredScreenName
						unless $self->can($DesiredScreenMethod);
						
					my $name = $self->$DesiredScreenMethod() || $vars->{screenname};
					# $self->trace("'$DesiredScreenMethod' returned '$name'");
					return $name if $name eq $DesiredScreenName;
				
				# We've been redirected to $name. Perform setup for that.
				
					my $nameScreenMethod = "${name}_screen";
					return $name
						unless $self->can($nameScreenMethod);
						
					my $name2 = $self->$nameScreenMethod();
					# $self->trace("'$nameScreenMethod' returned '$name2'");
					return $name if $name eq $name2;
					
				# We couldn't setup our redirection screen, fall back to starting a new session.
					
					$name = $self->StartNewSession;
					# $self->trace("StartNewSession returned '$name'");
					return $name || 'index';
			};
			
		! $@
			|| die "Syntax error in job->DoApplication loadable screen:\n$@";
			
		$vars->{screenname} = $DisplayableScreenName;
		
	# Announce new screen to trace.
	
		my $banner = '*' x 80;
		# $self->trace
		#	("\n\n\t$banner\n\t\t\t Next screen = '$DisplayableScreenName' \n\t$banner\n\n");
		
	# Print out standard headers.
	
		print $query->header( -Pragma => 'no-cache' );
		print '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">'."\n";
		
	# Interpret the screen.
	
		my $ScreenFile = "$DisplayableScreenName.ezscr";
		
		while ($ScreenFile)
		{
			$Interpreter->InterpretFile($ScreenFile);
			$ScreenFile = $Interpreter->{RedirectFile};
			( $vars->{screenname} = $1 )
				if $ScreenFile =~ /^(.+).ezscr$/s;
		};
		
		
	# Print either that or our error screen.
	
		print $self->{Died} ? $self->DieScreen : $Interpreter->{OutputBuffer};
}	
#-----------------------------------------------------------------------------------------------------#
sub DieScreen
{
	my $self 				= shift;
	my $Interpreter = $self->{Interpreter};
	my $DieMessage 	= $self->{DieMessage};
	
	my $FileName = $Interpreter->{FileName};
	$FileName =~ s/ezscr/screen/s;
	
	my ( $DispatcherDieScreen, $ReadErrorMessage );
	$self->ReadEntireFile( 'ezApp/ParseError.html', \$DispatcherDieScreen, \$ReadErrorMessage )
		|| die "Could not read 'ezApp/ParseError.html': $ReadErrorMessage\n";
	$DispatcherDieScreen =~ s/\$ErrorMessage/$DieMessage/sg;
	$DispatcherDieScreen =~ s/\$FileName/$FileName/sg;
	$DispatcherDieScreen =~ s/\$LineNumber/$Interpreter->{LineNumber}/sg;
$DispatcherDieScreen }
#-----------------------------------------------------------------------------------------------------#
sub	LoadButtonAttributes				# Recover and interpret attributes of last ezbutton hit.
																# Remove all button attributes from $vars.
																# The vars key for the button itself is:		
																#			"$ObjectName:ezbutton:$ButtonName:$seqno"
																# The vars key for each attribute is:
																#			"$ObjectName:ezbutton:$ButtonName:$seqno:$AttrName"
{
	my ($self, $varScanString) 		= @_;
	my $vars 											= $self->{vars};
	my $PersistentVars						= $self->{PersistentVars};
	
	return unless $varScanString 	=~ /\t([^\t:]*:ezbutton:([^\t:]+):[^\t:]+)\t/s;
	
	# An ezbutton was hit
		my $ValuesAttr 							= {};
		my $ExtendedName						= $1;
		$ValuesAttr->{name} 				= $2;
		$ValuesAttr->{label}				= $vars->{$1} ;
		
	# Reconstruct Attr from $vars.
		
			while ( $varScanString =~ /\t($ExtendedName:([^\t]+))(?=\t)/sg )
				{ $ValuesAttr->{$2} = $vars->{$1} };
		
	# Trace.
	
		$self->tracedump("pars recovered from last button:\n\t", $ValuesAttr)
			if $self->{fhTrace};
			
		# Store our privates (except for persistentvars & nonpersistentvars) 
		# in $self and delete all privates from ValuesAttr.  This allows the screen logic
		# to know all about the last button the user hit, including its label, TagText, etc.
		# Hash values which remain will be used to initialize the next screen's pars.
		
			@$self{qw(LastHitButtonScreen LastHitButtonName LastHitButtonLabel LastHitButtonTagText)}
				=@$ValuesAttr{ qw(screen name label tagtext) };
				
			foreach (@ezbuttonAttrVarNames)
				{ delete $ValuesAttr->{$_} }
				
		# Delete all button params from $vars.
		
			while ( $varScanString =~ /\t([^:\t]*:ezbutton:[^:\t]+(:[^\t]+)?)(?=\t)/sg )
			{
				delete $vars->{$1}
			};
			
		# Move rest of $ValuesAttr to $vars for our next screen's pars, and make it persistent.
		
			foreach (keys %$ValuesAttr)
			{
				my $varname = "ezpar:$_";
				$vars->{$varname} = $ValuesAttr->{$_};
				$PersistentVars->{$varname}++
			}
$ValuesAttr}
#-----------------------------------------------------------------------------------------------------#
sub CompileScreensIfOutOfDate
{
	my $self = shift;
	
	my $SourceIndex 		= 'screens/index.scr.txt';
	my $CompiledIndex		= 'ezscrs/index.ezscr';
	
	if ( ! -e $SourceIndex )
	{
		die "Cannot find '$SourceIndex'";
	};
	
	if ( ! -d 'ezscrs' )
	{
		die "Cannot find 'ezscrs/'";
	};
	
	if ( ! -e $CompiledIndex )
	{
		$self->CompileScreens;
		return;
	};
	
	
	# Compute the latest modification time ($Max_mtime) for our source files.
	
		my $dir 			= 'screens';
		my $Max_mtime = 0;
			
		(opendir DIR, $dir )
			|| die "Could not open '$dir' as a directory";
			
		my @files = readdir DIR;
		@files
			|| die "Directory '$dir' is empty";
			
		foreach (@files) 
		{ 
			$_ =~ /^.+?\.txt$/s
				or next;
				
			my $mtime = (stat "screens/$_")[9];
			$Max_mtime = $mtime
				if $mtime > $Max_mtime;
				
			# die "file='$_', mtime='$mtime', Max_mtime='$Max_mtime'\n";
		};
			
			
	# Compare $Max_mtime to modification time of 'ezscrs/index.ezscr'.
	
		my $Compile_mtime = (stat $CompiledIndex)[9];
		$self->CompileScreens
			if $Max_mtime > $Compile_mtime;
}
#-----------------------------------------------------------------------------------------------------#
sub CompileScreens		# Compile our screens.
{
	my $self = shift;
	
	$self->Note('Recompiling screens');

	eval "use ezApp::Compiler";
	
	!$@
		|| die "Could not use ezApp::Compiler: $@";
	
	my $Compiler = ezApp::Compiler->new;
	
	my $dir = 'screens';
		
	(opendir DIR, $dir )
		|| die "Could not open $dir as a directory";
		
	my @files = readdir DIR;
	@files
		|| die "Directory '$dir' is empty";
		
	foreach (@files) 
	{ 
		$_ =~ /^.+?\.scr.txt$/s
			|| next;
			
		$Compiler->CompileFile($_);
	};
}
#-----------------------------------------------------------------------------------------------------#
#
#		Utility methods
#
#----------------------------------------------------------------------
sub trace
{
	my ($self, $String) = @_;
	my $vars = $self->{vars};
	my $fhTrace = $self->{fhTrace}
		|| return;
	print $fhTrace "$String\n";
	if ( $vars->{URLFlushEachTrace} )
	{ 
		close $fhTrace; 
		(open $fhTrace, '>>ParseTrace.txt' )
			|| return $self->dieClean("Could not reopen >>ParseTrace.txt: $!")
	}
}
#-----------------------------------------------------------------------------------------------------#
sub tracedump
{
	my $self = shift;
	# $self->trace( $self->dump(@_) )
}
#-----------------------------------------------------------------------------------------------------#
sub dumpHTML
{
	my $self = shift;
	my $dump = HTML::Entities::encode_entities( $self->dump(@_) );
	$dump =~ s/\n/<br>\n/sg;
$dump}
#-----------------------------------------------------------------------------------------------------#
sub dump		# Return string showing values of arguments, one per line, with lines after the first indented.
						# Only unwrap refs to one level.  Prefix string with "\ndump - "
{
	my ( $self, @variables ) = @_;
	my $dump = "\ndump - ";
	my $refvalue;
	
	foreach my $variable (@variables)
	{
		if (! ( $refvalue = ref $variable ) )
			{ $dump .= $self->dumpSimple($variable) }
		else
		{
			$dump .= "$refvalue";
			if ($refvalue =~ /REF|SCALAR/s )
			{
				$dump .= ' to ' . $self->dumpSimple($$variable)
			}
			elsif ($refvalue =~ /CODE|GLOB/s )
				{}
			elsif ($refvalue eq 'ARRAY' )
			{
				foreach (@$variable)
				{
					$dump .= "\n\t\t".$self->dumpSimple($_)
				}
			}
			elsif ($refvalue eq 'HASH' )
			{
				foreach (keys %$variable)
				{
					$dump .= "\n\t\t$_ -- ".$self->dumpSimple($variable->{$_})
				}
			}
		};
		
		$dump .= "\n\t";
	};
	
substr $dump,0,-1 }
#-----------------------------------------------------------------------------------------------------#
sub dumpSimple
{
	my ($self, $variable) = @_;
	my ($dump, $refvalue) ;
	
	if (!defined $variable)
		{ $dump = 'undef' }
	elsif ($refvalue = ref $variable)
		{ $dump = "ref to $refvalue" }
	else
		{ $dump = "'$variable'" }
	
$dump}
#-----------------------------------------------------------------------------------------------------#
sub ReadEntireFile			# Read file into ContentsRef and return 1.
												# If problems, return '' and set ErrorMessage.
{	
	my ( $self, $FileName, $ContentsRef, $ErrorMessageRef ) = @_  ;
	if ( ! open IN, "<$FileName" ) {
		$$ErrorMessageRef = "Cannot open $FileName for input: $!" ;
		return '' ;
	} ;
	my $OldSep = $/ ; undef $/ ; 				## Mask $/ to read in one slurp
		$$ContentsRef = <IN> ;
	undef $/ ; $/ = $OldSep ;						## Restore record terminator and correct BBEdit sytax coloring.
	close IN ;
1 } 
#-----------------------------------------------------------------------------------------------------#
sub Note
{
	my ($self, $NoteText, $NotesOrWarnings) = @_;
	my $html = HTML::Entities::encode_entities($NoteText);
	$html =~ s/\n/<br>\n/sg;
	$html =~ s/ /\&nbsp;/sg;
	$html =~ s/\t/\&nbsp;\&nbsp;\&nbsp;\&nbsp;/sg;
	$self->htmlNote($html, $NotesOrWarnings);
1}
#-----------------------------------------------------------------------------------------------------#
sub htmlNote
{
	my ($self, $html, $NotesOrWarnings) = @_;
	$NotesOrWarnings ||= 'jobnotes';
	my $vars = $self->{vars};
	$html &&= "$html.";
	$vars->{$NotesOrWarnings} = join "<br>\n", $vars->{$NotesOrWarnings} || (), $html || ();
1}
#-----------------------------------------------------------------------------------------------------#
sub Warn
{
	my ($self, $text) = @_;
	$self->Note($text, 'jobwarnings');
''}
#-----------------------------------------------------------------------------------------------------#
sub htmlWarn
{
	my ($self, $html) = @_;
	$self->htmlNote($html, 'jobwarnings');
''}
#-----------------------------------------------------------------------------------------------------#
#
#		Methods which help manage tagged paramater lists.
#
#-----------------------------------------------------------------------------------------------------#
my $NamedArgPattern = '^-([A-Za-z^][A-Za-z0-9^]*)$' ;
#-----------------------------------------------------------------------------------------------------#
sub RecordInheritDefaults					# Return DefaultRecord, except where overridden by Record.
																	# Don't overwrite where Record value is 'undef'.
{
	my ( $self, $Record, $DefaultRecord ) = @_ ;
	return $Record unless $DefaultRecord ;								# If default missing, just use Record.
	my $NewRecord = { %$DefaultRecord } ;									# Make a copy of the default -- don't mess up caller's copy.
	return $NewRecord unless $Record ;										# If Record missing, just use default.
	foreach ( keys %$Record ) {
		$NewRecord->{$_} = $Record->{$_}										# Overwrite with each existing field from Record,
			if ( defined $Record->{$_} ) 											# unless the value is 'undef' and a value already exists.
				or ! exists $NewRecord->{$_} ;	
	} ;
$NewRecord}
#-----------------------------------------------------------------------------------------------------#
sub AddArgsToRecord								# Add (or overwrite) entries { name=>value ... } in $Record,
																	# computed from @_ = ( -name=>value ... ).
																	# value = 'undef' is processed like any other value.
																	# Return the resulting record
{
	my ( $self, $Record ) = ( shift(), shift() ) ;
	while ( @_ ) {
		my ( $_ArgName, $Arg ) = ( shift(), shift() ) ;
		$_ArgName =~ /$NamedArgPattern/s 
			|| return $self->dieClean
				("job->AddArgsToRecord found bad ArgName '$_ArgName'");
		$Record->{$1} = $Arg ;
	} ;
$Record }
#-----------------------------------------------------------------------------------------------------#
sub HashAsArgs 				# Convert hash into named list: ( -key1=>value1, ... )
											# Undefined value generates ( -key, ... )  ( eg useful for '-readonly' )
{																	
	my ( $self, $Hash ) = @_ ;																	
	return () unless 'HASH' eq ref $Hash ;
	my @NamedList = map { 
		my $Value = $Hash->{$_} ;
		defined $Value
			? ( "-$_" => $Hash->{$_} )
			: ( "-$_" )
	} keys %$Hash ; 
@NamedList}
#-----------------------------------------------------------------------------------------------------#
sub ArgsAsHash  				# Unpack argument list and assign defaults.
{
	my ( $self, $Defaults ) = 
		( shift() , shift() ) ;

	# Our calling scenario looks like this:
	#
	#	sub Caller { my $self = shift ;
	#		my $Defaults = [ -name=>'textfield', -size=>20 ] ;	# Use these values where none supplied in @_.
	#		my $ArgRecord = $job->ArgsAsHash( $Defaults, @_ ) ;
	#
	# 
	# @_ can begin with (value, value, ...) format
	# and switch at any point to (-name=>value, ...) format, but it can't 
	# switch back.  If value args occur, they are assigned to
	# sequential names from $Defaults, beginning with the first.
	# There must be enough defaults to cover all value args.
	# 
	# Return a record (hash) of { name=>value, ... } based on Defaults,
	# as overridden by non-undef values from Callers argument list.
	#
	
	my @Defaults = @$Defaults ;
	my $DefaultRecord = $self->AddArgsToRecord( {}, @Defaults ) 
		|| return undef;
	
	my $CallingRecord = {} ;
	my ( $_ArgName, $Arg ) ;
	
	# First pick up any (value) args.
		while 	( 
							@_ && 
							(																													# We have another arg.
								ref ( $Arg = @_->[0] ) ||																# Is it a non-scalar?
								$Arg !~ /$NamedArgPattern/s															# or a non '-name'?
							)
						)
		{																																		# Yes, we have a (value) arg.
			shift ;																														# Trim the arg from @_.				
			$_ArgName = shift @Defaults 			  															# Pick up its name from Defaults.
				|| return $self->dieClean
					("job->ArgsAsHash has no name left for value '$Arg'") ;
			shift @Defaults ;																									# Skip past the default value.
			$_ArgName =~ /$NamedArgPattern/s 																	# Get the name without '-'.
				|| return $self->dieClean
					("Bad arg name '$_ArgName' to job->ArgsAsHash") ;
			$CallingRecord->{$1} = $Arg ;
		} ;
		
	# Next, pick up any (-name=>value) args.
		$self->AddArgsToRecord( $CallingRecord, @_ ) 
			|| return undef;
		
	# Finally, observe proper inheritance.
		my $ArgRecord = $self->RecordInheritDefaults( $CallingRecord, $DefaultRecord ) ;
		
$ArgRecord }
#-----------------------------------------------------------------------------------------------------------
sub SendMessage 					# Send the message, which includes its headers.
													# Return '' if success, else a diagnostic message.
													# Date header is added automatically by unix.
													# We look first for 'qmail', then 'sendmail'.
{ 
	my $self = shift;
	my $MessageTextRef = shift ;													
	use IO::File ;
	my $QSendMail = '/var/qmail/bin/qmail-inject' ;
	my $Sendmail = '/usr/sbin/sendmail';
	if ( -e $QSendMail && -x $QSendMail ) {
		my $Pipe = IO::File->new("|$QSendMail")
			or return "Failed to open pipe to qmail-inject: $!" ;
		print $Pipe $$MessageTextRef;
		$Pipe->close();
		return "Failed to send e-mail message: $@"
			unless (($?>>8) == 0) ;
	} elsif ( -e $Sendmail && -x $Sendmail ) {
		open(SENDMAIL, "|$Sendmail -oi -t -odq") 
			or return "Cannot open $Sendmail: $!";
		print SENDMAIL $$MessageTextRef ;
		close(SENDMAIL)
			or return "Cannot close $Sendmail: $!";
	} else {
		return 'Could not find a way to send mail'
	}
''}
#-----------------------------------------------------------------------------------------------------#
1;
