#!/usr/bin/perl -w
use diagnostics;
use strict;

package job;  
    use ezApp::job; 
    use vars qw(@ISA);  
    @ISA = qw(ezApp::job);

use ccKernel;
use ccConfig;
  
#--------------------------

my (
  $SuperUser, $SuperUserPassword, $MaxCurrenciesOwned,
  $ccVersion, $Disclaimer, $AdminTradeMessageString,
  $PasswordMessage, $DBHandler, $DBName, $DBUser, $DBPassword
  );

my $EmailAddrPattern    = '^[^@]+@[^.@]+\\.[^@]+$';

my $OriginColumns  = 
  [ qw( AcctId AcctPassword AcctEmail AcctName LastLoginDate ) ]; 

my $AcksColumns  = 
  [ qw( AckDateTime SourceAcctId DestAcctId AckCurrencyId
  AckAmount AckItem SourceBalance DestBalance ) ];
  
my $LedgerColumns = 
  [ qw( AcctId CurrencyId Balance TradesTotal TradesCount
  LastTradeDate ) ];

#-----------------------------------------------------------

  # Human readable names for screen states.
  
      my $ReadableScreenNames = 
      {
        index               => 'login',
        SuperMain           => 'super_main',
        SuperShowUsers      => 'show_users',
        SuperShowCurrencies => 'show_currencies',
        UserMain            => 'main',
        ShowTrades          => 'view_trades',
        UserCurrencies      => 'view_currencies',
        AdminCurrency       => 'admin_currency',
        UpdateAccount       => 'update_account'
      };
    
  # vars to clear when Backing from a screen state.
  
      my $ScreenBackoutVars = 
      {
        SuperMain => 
    [ qw (  ActiveAcctId ActivePassword SendToAddress NewAcctId
    NewAcctName NewAcctEmail ActiveAcctName ActiveAcctEmail
    TradeItem TradeWith TradeCurrency TradeAmount AdminTradeItem
    AdminTradeInitiator AdminTradeWith AdminTradeCurrency
    AdminTradeAmount ) ] ,
        SuperShowUsers => 
            [ qw ( ) ] ,
        SuperShowCurrencies => 
            [ qw ( ) ] ,
        UserMain  => 
    [ qw (  ActiveAcctId ActivePassword SendToAddress NewAcctId
    NewAcctName NewAcctEmail ActiveAcctName ActiveAcctEmail
    TradeItem TradeWith TradeCurrency TradeAmount AdminTradeItem
    AdminTradeInitiator AdminTradeWith AdminTradeCurrency
    AdminTradeAmount ) ] ,
        UpdateAccount => 
            [ qw ( UpdateAcctName UpdateAcctEmail ) ] ,
        UserCurrencies => 
    [ qw ( RegisterForCurrencyId NewCurrencyId CurrencyIdToAdmin
    ) ] ,
        ShowTrades => 
            [ qw ( ) ] ,
        AdminCurrency => [ qw ( AdminAcctId ) ] 
      };
  
#----------------------------------------------------------
sub new 
{
  my $class                         = shift;
  my $self                          = ezApp::job->new;
  bless $self, $class;
  $self->{ScreenBackoutVars}        = $ScreenBackoutVars;
  $self->{AcksColumns}              = $AcksColumns;
  $self->{LedgerColumns}            = $LedgerColumns;

  ( $SuperUser, $SuperUserPassword, $MaxCurrenciesOwned,
  $ccVersion, $Disclaimer, $AdminTradeMessageString,
  $PasswordMessage, $DBHandler, $DBName, $DBUser, $DBPassword
  ) = &ccConfig::SetConfig;
  
# die "job.pm new: '$DBHandler','$DBName','$DBUser','$DBPassword'\n";
  
  my $ccKernel = $self->{ccKernel}  = 
    ccKernel->new( $self, $DBHandler, $DBName, $DBUser,
    $DBPassword ) ;
    
  $self->{ccDB}                     = $ccKernel->{ccDB};
  $self->{isopen}                   = $ccKernel->{isopen} ;
  $self->{NotOpenMessage}           = 
      "Database $DBName could not be opened";
$self}
#-------------------------------------------------------------
sub StartNewSessionSetup    # Initialize a new sesssion.
{
  my $self                  = shift;
  my $vars                  = $self->{vars};
  my $PersistentVars        = $self->{PersistentVars};

  $vars->{ShowTradesCount}  = '5';
  $vars->{ccVersion}        = $ccVersion;
  $vars->{disclaimer}       = $Disclaimer;
  $PersistentVars->{ShowTradesCount}++;
  $PersistentVars->{ccVersion}++;
  $PersistentVars->{disclaimer}++;
  $PersistentVars->{GoBackStack}++;
  
'index'}
#-------------------------------------------------------------
sub ContinueSessionSetup  
                          
{
  my $self = shift;
  my $vars = $self->{vars};
  
  return '' if $vars->{screenname} eq 'index';
  $self->LoginError 
    && return 'index';
  
  $self->LoadAcctData;
    
''}
#---------------------------------------------------------------
sub LoadAcctData
{
  my $self            = shift;
  my $vars            = $self->{vars};
  my $PersistentVars  = $self->{PersistentVars};
  my $ccDB            = $self->{ccDB};
  my $ccKernel        = $self->{ccKernel};
  my $AcctId          = $vars->{ActiveAcctId};
  
  my $CurrencyOwnerIds = $ccKernel->{CurrencyOwnerIds} ;
  
  if ($AcctId eq $SuperUser) 
  { 
    # Mark super user as owning all currencies.
      $vars->{OwnedCurrencyIds} = [ keys %$CurrencyOwnerIds ];
    return 1;
  };
  
  ## Read in user's origin record.
    
    my $Records = $ccDB->Records( 'Origins',
      -Columns  =>  ['AcctEmail', 'AcctName'],
      -Where    =>  "AcctId='$AcctId'"
    );
    
    $Records && @$Records
      || return $self->Warn
        ("Could not read account record for $AcctId");
      
    my $Record = shift @$Records;
      
    @$vars{'ActiveAcctEmail', 'ActiveAcctName'} = 
      @$Record{'AcctEmail', 'AcctName'};
    
    $PersistentVars->{ActiveAcctEmail}++;
    $PersistentVars->{ActiveAcctName}++;
    
  
    $vars->{AcctLedgers} = $ccDB->Records('Ledgers',
      -Columns  => $LedgerColumns,
      -Where    => "AcctId='$AcctId'",
      -OrderBy  => 'LastTradeDate desc'
    ) || return $self->Warn( "Could not read ledger records" );
      
    $vars->{RegisteredCurrencyIds} = 
      [ sort map { $_->{CurrencyId} } 
        @{ $vars->{AcctLedgers} } ];
    

  ## Setup array of owned currencies.
    
    my $OwnedCurrencyIds = [];
    
    foreach (keys %$CurrencyOwnerIds)
      { ( push @$OwnedCurrencyIds, $_ ) 
        if $AcctId eq $CurrencyOwnerIds->{$_} };
      
    $vars->{OwnedCurrencyIds} = $OwnedCurrencyIds;
1}
#----------------------------------------------------------
sub Logout      # Logout = return to blank login screen.
{
  my ( $self, $Warning ) = @_;
  my $vars = $self->{vars};
  foreach my $ScreenState ( keys %$ScreenBackoutVars ) 
  {
    delete @$vars{ @{ $ScreenBackoutVars->{$ScreenState} } };
  };
  delete $vars->{GoBackStack};
  $self->Warn($Warning);
'index'}
#----------------------------------------------------------
sub Refresh_button        # Return to same screen
{
  my $self = shift;
  $self->Note('Screen refreshed.  No transaction performed');
'' }
#------------------------------------------------------------
sub GoBack_button     # Return to previous screen.
{ 
  my $self            = shift;
  my $vars            = $self->{vars};
  my $PersistentVars  = $self->{PersistentVars};
  
  # Delete variables from current screen.
  
    my $ScreenName = $vars->{screenname};
    delete @$vars{ @{ $ScreenBackoutVars->{$ScreenName} } };
    delete @$PersistentVars{ @{
      $ScreenBackoutVars->{$ScreenName} } };
    
  # Pop our parent screen state from stack. 
  
    my @GoBackStack = split '/', $vars->{GoBackStack};
    ( my $GoBackScreen = pop @GoBackStack )
      or return $self->Warn('GoBack stack empty');
    $vars->{GoBackStack} = join '/', @GoBackStack;
    
$GoBackScreen}
#------------------------------------------------------------
sub RegisterUser_button   # Register a new user.
{
  my $self = shift;   
  
  $self->{isopen}
    || return $self->Warn( $self->{NotOpenMessage} );
  
  my ($vars, $ccDB, $ccKernel) = @$self
    { qw(vars ccDB ccKernel) };
    
  my ( $NewAcctId, $NewAcctName, $NewAcctEmail ) = 
    @$vars{ qw( NewAcctId NewAcctName NewAcctEmail ) };
    
  $NewAcctId && $NewAcctName && $NewAcctEmail
    || return $self->Warn
      ('All fields must be filled in '.
        'for new account registration');
    
  $NewAcctId = lc $NewAcctId;
  
  ! $self->{UserPasswords}->{$NewAcctId}
    || return $self->Warn
      ("$NewAcctId is already a registered user");
  
  $NewAcctName =~ tr/\"/\'/;    
  
  $NewAcctEmail =~ /$EmailAddrPattern/
    || return $self->Warn
      ("'$NewAcctEmail' not a valid email address");
    
  my $NewPassword = $self->NewPassword;
  
  $ccKernel->AddOrigin( $NewAcctId, $NewPassword )
    || return $self->Warn
      ("Could not create account record for $NewAcctId");
    
  $ccDB->UpdateRecords('Origins',
    -Columns  => [ 'AcctEmail', 'AcctName' ],
    -Values   => [ $ccDB->QuotedSQLvalue($NewAcctEmail),  
      $ccDB->QuotedSQLvalue($NewAcctName) ],
    -Where    => "AcctId='$NewAcctId'"
  ) || return $self->Warn
    ("Your registration record is incomplete");

  $self->Note("You are now registered as $NewAcctId");
  
  $ccKernel->FetchPrimaryHashes   # && $self->LoadAcctData
    || return '';

  $self->SendPasswordMessage
    ( $NewAcctId, $NewPassword, $NewAcctEmail, $NewAcctName )
    || return '';
    
'index'}
#-----------------------------------------------------------
sub NewPassword
{ 
  my $self = shift;
  
  my $Password = rand;
  $Password =~ tr/.//d;
  $Password = substr( $Password, -7, 5 );
  
$Password}
#----------------------------------------------------------
sub Login_button    # Login an existing user.
{ 
  my $self            = shift;
  my $vars            = $self->{vars};
  my $job             = $self->{job};
  my $ccDB            = $self->{ccDB};
  my $PersistentVars  = $self->{PersistentVars};
  
  my ( $ActiveAcctId, $ActivePassword ) 
    = @$vars{ qw( ActiveAcctId ActivePassword ) }
      = map {lc $_} 
        @$vars{ qw( ActiveAcctId ActivePassword ) };
    
  $ActiveAcctId && $ActivePassword
    || return $self->Warn
      ('ID and Password both required for login');
  
  my $error;
  ( $error = $self->LoginError )
    && return $self->Warn($error);
    
  if ($ActiveAcctId ne $SuperUser) 
  {
    $ccDB->UpdateRecords( 'Origins',
      -Columns  => ['LastLoginDate'],
      -Values   => ['now()'],
      -Where    => "AcctId = '$ActiveAcctId'"
      ) or return '';
  };
  
  $self->LoadAcctData
    || return '';
  
  $self->PushGoBackStack;
  
# die "ActiveAcctId=$ActiveAcctId";
  
  return $ActiveAcctId eq $SuperUser
    ? 'SuperMain'
    : 'UserMain';
}
#---------------------------------------------------
sub SuperShowUsers_button   # Go to SuperShowUsers.
{
  my $self = shift;
  $self->PushGoBackStack;
'SuperShowUsers'}
#------------------------------------------------------
sub SuperShowUsers_screen   # Load data SuperShowUsers.
{
  my $self = shift;
  my ( $vars, $ccDB, $ccKernel ) = 
    @$self{ qw( vars ccDB ccKernel ) };
    
  my $CurrencyOwnerIds = $ccKernel->{CurrencyOwnerIds} ;
  my $CurrencyOwners = 
    { map { $CurrencyOwnerIds->{$_} => 1 } 
      keys %$CurrencyOwnerIds };
  $vars->{CurrencyOwners} = $CurrencyOwners;
  
  my $Origins = $ccDB->Records( "Origins", 
    -Columns  => $OriginColumns,
    -OrderBy  => "AcctId"
  );
  
  $Origins
    || return $self->Warn
      ('Could not read Origins records');
    
  $vars->{Origins} = $Origins;
  
'SuperShowUsers'}
#--------------------------------------------
sub SuperShowCurrencies_button    
{
  my $self = shift;
  $self->PushGoBackStack;
'SuperShowCurrencies'}
#---------------------------------------------
sub SuperShowCurrencies_screen  
{
  my $self = shift;
  my ( $vars, $ccDB, $ccKernel ) = 
    @$self{ qw( vars ccDB ccKernel ) };
    
  $vars->{CurrencyOwnerIds} = 
    $ccKernel->{CurrencyOwnerIds} ;
  
'SuperShowCurrencies'}
#---------------------------------------------
sub UpdateAccount_button    
{
  my $self = shift;
  my $vars = $self->{vars};
  
  $self->{isopen}
    || return $self->Warn( $self->{NotOpenMessage} );
  
  $self->Login_button 
    || return '';

  @$vars{ qw( UpdateAcctEmail UpdateAcctName ) } = 
    @$vars{ qw( ActiveAcctEmail ActiveAcctName ) } ;
    
'UpdateAccount'}
#-------------------------------------------------
sub PostAcctUpdates_button    # Post account updates.
{
  my $self = shift;
  
  my ( $vars,  $ccDB, $ccKernel ) = 
    @$self{ qw ( vars ccDB ccKernel ) };
  
  my ( $AcctId, $OldPassword, $OldEmail, $OldName, $NewName,
    $NewEmail ) = @$vars{ qw( ActiveAcctId ActivePassword
    ActiveAcctEmail ActiveAcctName UpdateAcctName
    UpdateAcctEmail ) };
    
  ($NewName ne $OldName) || ($NewEmail ne $OldEmail)
    || return $self->Warn
      ('Values were the same: nothing posted');
    
  $NewEmail =~ /$EmailAddrPattern/
    || return $self->Warn
      ("'$NewEmail' is not a valid email address");

  my $NewPassword = $NewEmail eq $OldEmail
    ? $OldPassword
    : $self->NewPassword;

  $ccDB->UpdateRecords( 'Origins',
    -Columns  => [ qw( AcctName AcctEmail AcctPassword ) ],
    -Values => [ map { "'$_'" } 
      $NewName, $NewEmail, $NewPassword  ],
    -Where  => " AcctId='$AcctId' "
  ) or return '';
  
  $self->Note("Account updated");
  
  if ( $NewEmail ne $OldEmail ) 
  {
    # Send messages.  If error occurs, we proceed anyway.  
    # We may need to implement roll back for this. *note*
    
      $self->SendPasswordMessage
        ( $AcctId, $NewPassword, $NewEmail, $NewName );
      $self->SendPasswordMessage
        ( $AcctId, $NewPassword, $OldEmail, $NewName, $NewEmail,
        $OldName ) ;
  };
  
  $vars->{ActiveAcctName}   = $NewName;
  $vars->{ActivePassword}   = $NewPassword;
  $vars->{ActiveAcctEmail}  = $NewEmail;
    
  $ccKernel->FetchPrimaryHashes && $self->LoadAcctData
    || return '';
  
'UserMain'}
#--------------------------------------------------------
sub EmailPassword_button  
{
  my $self = shift;
  
  my ( $vars, $ccDB, $ccKernel ) = 
    @$self{ qw( vars ccDB ccKernel ) };
  
  my ( $AcctId, $SendToAddress ) = 
    @$vars{ qw( ActiveAcctId SendToAddress ) };
  $AcctId = lc $AcctId;
  
  ( $AcctId && $SendToAddress ) 
    || return $self->Warn
      ('ID and Email address both required to send Password');

  my $OriginRecords = $ccDB->Records('Origins', 
      -Columns  => [ qw( AcctPassword AcctEmail AcctName ) ],
      -Where    => "AcctId='$AcctId'" );
  
  $OriginRecords && @$OriginRecords 
    || return $self->Warn('Could not read account record');
    
  my $OriginRecord = shift @$OriginRecords;
  
  my ( $AcctPassword, $AcctEmail, $AcctName ) = 
    @$OriginRecord{ qw ( AcctPassword AcctEmail AcctName ) };
  
  $AcctEmail eq $SendToAddress
    || return $self->Warn
      ('The Email address you specified does not match '.
      'your registered address... message not sent');
    
  $self->SendPasswordMessage
    ( $AcctId, $AcctPassword, $AcctEmail, $AcctName );
    
''}
#-----------------------------------------------
#-----------------------------------------------
sub GoToCurrencies_button
{
  my $self = shift;
  $self->PushGoBackStack;
'UserCurrencies'}
#-----------------------------------------------
sub MakeTrade_button        # User make a new trade.
{
  my $self = shift;
  my $vars = $self->{vars};
  
  my ( $Item, $DestAcctId, $CurrencyId, $Amount ) =
    @$vars{ qw( TradeItem TradeWith TradeCurrency TradeAmount ) };
    
  $self->Transfer
    ( 'trade', $Item, $vars->{ActiveAcctId}, 
      $DestAcctId, $CurrencyId, $Amount ) 
        || return '';
      
  delete @$vars{ qw( TradeItem TradeWith TradeAmount ) };
''}
#-----------------------------------------------
#-----------------------------------------------
sub AdminTrade_button   # Currency owner trade or reversal.
{
  my $self              = shift;
  my $vars              = $self->{vars};
  my $ccDB              = $self->{ccDB};
  my $AdminAcctId       = $vars->{ActiveAcctId};
  my $TransferDirection = $vars->{'ezpar:transferdirection'};
  
  my ( $Item, $SourceAcctId, $DestAcctId, $CurrencyId, $Amount ) =
    @$vars{ qw(AdminTradeItem AdminTradeInitiator AdminTradeWith
    AdminTradeCurrency AdminTradeAmount) };
    
  $Item = "*!*$Item";   # Mark as admin trade.
    
  $self->Transfer
    ( $TransferDirection, $Item, $SourceAcctId, 
      $DestAcctId, $CurrencyId, $Amount ) 
        || return '';
      
  delete @$vars{ qw( AdminTradeItem AdminTradeInitiator 
    AdminTradeWith AdminTradeAmount ) };

  ## Pickup info on trade parties and send them each a message.
  
    my $TradeType = $TransferDirection eq 'trade'
      ? 'trade'
      : 'trade reversal';

    foreach ($SourceAcctId, $DestAcctId)
    {
      my $Rows = $ccDB->Records( 'Origins',
        -Columns  => ['AcctName','AcctEmail'],
        -Where    => "AcctId='$_'"
      );
      
      $Rows && @$Rows
        || return $self->Warn("Could not read account data for $_");
      
      my $AcctRecord  = shift @$Rows;
      my $ToName      = $AcctRecord->{AcctName};
      my $ToEmail     = $AcctRecord->{AcctEmail};
      $Item           =~ s/\*!\*//s;
      my $Message     = eval($AdminTradeMessageString);
      
      $self->SendMessageUnlessNoEmail($Message, $ToEmail)
        || return '';
        
    };
''}
#-----------------------------------------------
sub FetchMyTrades_button      
{
  my $self = shift;
  my $vars                      = $self->{vars};
  
  $vars->{AdminTradeCurrency}   = $vars->{"ezpar:currencyid"};
  
  $self->PushGoBackStack;
'ShowTrades'}
#-----------------------------------------------
sub RegisterForCurrency_button    
{
  my $self = shift;
  
  my ( $vars, $ccKernel ) = @$self{ qw
    ( vars ccKernel ) };
    
  my ( $AcctId, $CurrencyId ) = @$vars{ 
    qw( ActiveAcctId RegisterForCurrencyId ) };
  
  $CurrencyId = lc $CurrencyId;
  
  return $self->Warn('No currency ID specified')
    unless $CurrencyId;
    
  $ccKernel->{CurrencyOwnerIds}->{$CurrencyId}
    || return $self->Warn
      ("$CurrencyId is not an established currency");
    
  $ccKernel->CreateLedger( $AcctId, $CurrencyId ) 
    || return '';
    
  $self->Note("You are now registered for $CurrencyId");
  
  delete $vars->{RegisterForCurrencyId};
  
  $self->LoadAcctData
    || return '';
  
'UserMain'}
#-----------------------------------------------
sub EstablishNewCurrency_button     # Add a new currency.
{ 
  my $self      = shift;
  
  my ( $vars, $ccDB, $ccKernel ) = @$self
    { qw( vars ccDB ccKernel ) };
    
  my ( $NewCurrencyId, $CreatorAcctId ) = map { lc $_ } @$vars 
    { qw( NewCurrencyId ActiveAcctId ) };
    
  $NewCurrencyId 
    || return $self->Warn('No currency ID specified');
  
  $self->IdsOK($NewCurrencyId) 
    || return '';
  
  
  ## Don't let one person establish too many currencies.
  
    my $CurrencyOwnerIds = $ccKernel->{CurrencyOwnerIds};
    my $OwnedCount = 0;
    
    foreach ( keys %$CurrencyOwnerIds )
      { $OwnedCount++ if $CreatorAcctId eq $CurrencyOwnerIds->{$_} };
  
    $OwnedCount < $MaxCurrenciesOwned
       || return $self->Warn
        ("Maximum of $MaxCurrenciesOwned currencies ".
          "may be established by one account");
  
  
  $ccKernel->AddCurrency($NewCurrencyId, $CreatorAcctId) 
    || return '';
  
  $ccDB->UpdateRecords('Currencies',
    -Columns  => ['CreatorAcctId'],
    -Values   => ["'$CreatorAcctId'"],
    -Where    => "CurrencyId='$NewCurrencyId'" 
  ) || return $self->Warn
    ("Could not finist writing new currency record");
  
  $self->Note
    ("$NewCurrencyId has been established by $CreatorAcctId.");
  delete $vars->{NewCurrencyId};
  
  $ccKernel->FetchPrimaryHashes && $self->LoadAcctData
    || return '';
  
'UserMain'}
#-----------------------------------------------
sub AdministerCurrency_button   # Administer an owned currency
{
  my $self = shift;
  $self->PushGoBackStack;
'AdminCurrency'}
#-----------------------------------------------
sub AdminCurrency_screen 
{
  my $self                = shift;
  my ( $vars, $ccDB, $ccKernel, $PersistentVars ) 
      = @$self{ qw( vars ccDB ccKernel PersistentVars ) };
                          
  $vars->{AdminAcctId}    = $vars->{ActiveAcctId};
  my $CurrencyIdToAdmin   = $vars->{CurrencyIdToAdmin};
  
  my $AdminLedgers = $vars->{AdminLedgers}  = 
    $ccDB->Records('Ledgers', 
      -Columns  => $LedgerColumns,
      -Where    => "CurrencyId='$CurrencyIdToAdmin'",
      -OrderBy  => 'AcctId'
  ) || return $self->Warn
    ("Could not read ledgers for $CurrencyIdToAdmin");
  
  ## Add a summary row with AcctId 'TOTAL',
    my $AdminSumRecord = 
      { AcctId => 'TOTAL', LastTradeDate => '' }; 
    
    foreach my $AdminLedger (@$AdminLedgers) {
      foreach ( qw (Balance TradesTotal TradesCount) ) {
        $AdminSumRecord->{$_} += $AdminLedger->{$_};
      };
      if ( $AdminSumRecord->{LastTradeDate} 
        lt $AdminLedger->{LastTradeDate} ) 
          { $AdminSumRecord->{LastTradeDate} =
          $AdminLedger->{LastTradeDate}; };
    };
    push @$AdminLedgers, $AdminSumRecord;
  
  $PersistentVars->{AdminAcctId}++;
  $PersistentVars->{CurrencyIdToAdmin}++;
  
'AdminCurrency'}
#-----------------------------------------------
sub SuperAdminCurrency_button   
{
  my $self = shift;
  my $vars = $self->{vars};
  $vars->{CurrencyIdToAdmin} = $vars->{"ezpar:currencyid"};
  $self->PushGoBackStack;
'AdminCurrency'}
#-----------------------------------------------
sub FetchHisTrades_button     
{
  my $self                      = shift;
  my $vars                      = $self->{vars};
  $vars->{AdminTradeCurrency}   = $vars->{"ezpar:currencyid"};
  $self->PushGoBackStack;
'ShowTrades'}
#-----------------------------------------------
#
#     Utility methods 
#
#-----------------------------------------------
sub IdsOK     # Verify validity of Ids.
{ 
  my ($self, @Ids ) = @_;
  
  foreach my $Id (@Ids) 
  {
    $Id =~ /^[a-z^][a-z0-9^]{2,}$/ 
      || return $self->Warn
        ("'$Id' not valid - should be at least three ".
          "letters and digits, with no spaces")
  }; 
  
1}
#-----------------------------------------------
sub PushGoBackStack     
{
  my $self = shift;
  my $vars = $self->{vars};
  my @GoBackStack = split '/', $vars->{GoBackStack};
  push @GoBackStack, $vars->{screenname};
  $vars->{GoBackStack} = join '/', @GoBackStack;
1}
#-----------------------------------------------
sub ScreenNameStack     
{
  my $self = shift;
  my $vars = $self->{vars};
  my @NameStack = map { $ReadableScreenNames->{$_} } 
    ( split '/', $vars->{GoBackStack} || '' ) , 
      $vars->{screenname};
join '>', @NameStack }
#-----------------------------------------------
sub LoginError    # Verify that AcctId and Password match
{ 
  my $self = shift;
  
  $self->{isopen}
    || return $self->{NotOpenMessage};
    
  my ( $vars, $ccKernel ) = 
    @$self{ qw( vars ccKernel ) };
    
  my ( $ActiveAcctId, $ActivePassword ) = map {lc $_} 
    @$vars{ qw( ActiveAcctId ActivePassword ) };
    
  return '' if
    ( $ActiveAcctId eq $SuperUser && $ActivePassword eq
    $SuperUserPassword );
  
  my $UserPasswords = $ccKernel->{UserPasswords};
  
  my $RealPassword = $UserPasswords->{$ActiveAcctId} 
    || return "'$ActiveAcctId' is not a valid account name";
    
  $RealPassword eq $ActivePassword 
    || return "Password incorrect, please try again";
    
''}
#-----------------------------------------------
sub LoginIsSuperUser    
{ 
  my $self = shift;
  my $vars = $self->{vars};
  
  my ( $ActiveAcctId, $ActivePassword ) = map {lc $_} 
    @$vars{ qw( ActiveAcctId ActivePassword ) };
    
  return $self->Warn('Password incorrect, try again') unless
    ( $ActiveAcctId eq $SuperUser ) 
      && $ActivePassword eq $SuperUserPassword;
    
1}
#-----------------------------------------------
sub ShowUserOrSuperMain     
{
  my $self = shift;
  return 
    $SuperUser eq lc $self->{vars}->{ActiveAcctId} 
      ? 'SuperMain' : 'UserMain' 
}                                                     
#-----------------------------------------------
sub Transfer      # Record or reverse a trade.
{
  my ( $self, $Direction, $Item, $SourceAcctId, $DestAcctId,
    $CurrencyId, $Amount ) = @_;
  
  my ($vars, $ccDB, $ccKernel) = 
    @$self{ qw( vars ccDB ccKernel ) };
    
  my $UserPasswords = $ccKernel->{UserPasswords};
  
  ($SourceAcctId, $DestAcctId, $CurrencyId) 
    = map { lc $_ } 
      $SourceAcctId, $DestAcctId, $CurrencyId;
    
  ($SourceAcctId && $DestAcctId)
    || return $self->Warn('Accounts must be filled in');
    
  $SourceAcctId ne $DestAcctId 
    || return $self->Warn
      ('An account cannot trade with itself');
  
  foreach($SourceAcctId, $DestAcctId)
  {
    $UserPasswords->{$_}
      || return $self->Warn("$_ is not a registered account");
  };
  
  $Amount =~ tr/.,//d;
  $Amount =~ /^[0-9]{1,8}$/s
    || return $self->Warn
      ('Amount must be non-negative number '.
        'with fewer than 9 digits');
  
  my $TradeAdjective = $Direction eq 'trade'
    ? 'posted'
    : 'reversed';
    
  $Direction eq 'reverse'
    && ( $Amount = - $Amount );

  $ccKernel->Transfer
    ($SourceAcctId, $DestAcctId, $CurrencyId, $Amount, $Item) 
    || return '';
  $self->Note("The trade has been $TradeAdjective");
  $self->LoadAcctData;
  
1}
#-----------------------------------------------
#-----------------------------------------------
sub SendPasswordMessage     
                            
{
  my ( $self, $AcctId, $AcctPassword, $AcctEmail, $AcctName,
    $NewEmail, $OldName ) = @_;
  my $vars = $self->{vars};
  
  my ( $ToAddr, $MailLine, $PasswordLine );
  
  if ($NewEmail) 
  {
    $ToAddr = $OldName;
    $MailLine = "New email address: $NewEmail";
    $PasswordLine = 'A new password was mailed to that address';
  } 
  else 
  {
    $ToAddr = $AcctName;
    $MailLine = "Email: $AcctEmail";
    $PasswordLine = "Password: $AcctPassword";
  };
    
  $self->SendMessageUnlessNoEmail
    ( eval($PasswordMessage), $AcctEmail )
      || return '';
1}
#-----------------------------------------------
sub SendMessageUnlessNoEmail
{
  my ($self, $Message, $EmailAddr)  = @_;
  my $vars = $self->{vars};
  
  if ( $vars->{URLNoEmail} )
    { $self->Note("$Message") }
  else
  {
    my $ErrMsg = $self->SendMessage(\$Message);
    ! $ErrMsg
      || return $self->Warn($ErrMsg); 
    $self->Note("A message has been sent to $EmailAddr");
  };
1}
#-----------------------------------------------
#
#     Emit HTML.
#
#
      my ( $WarningColor, $LtGreenColor ) = ( '#FF6666', '#66FF99' );
#
#-----------------------------------------------
sub EmitTradesDataRow         # Emit HTML for an Ack data row.
{
  my ( $self, $AcctId, $Record ) = @_;
  my $vars = $self->{vars};
  my ( $Date, $SourceAcctId, $DestAcctId, $CurrencyId,
    $Amount, $Item, $SourceBalance, $DestBalance ) = @$Record{
    qw( AckDateTime SourceAcctId DestAcctId AckCurrencyId AckAmount
    AckItem SourceBalance DestBalance ) };
      
  my ( $Deposit, $Withdrawal, $Balance, $SetVendorColor,
    $SetRvsColor ) = ( ('') x 10 );
  my $AbsAmount = abs $Amount;
  $Date = substr $Date, 0, 10;
  
  if ($Item =~ /^\*!\*/s)
  {
    $vars->{AdminTradeDisplayed}++;
    $Item = substr $Item, 3;
    $SetRvsColor = "bgcolor='$WarningColor'";
  };
  
  if ( $SourceAcctId eq $AcctId or $AcctId eq 'TOTAL') 
  {
      
      $Balance = $SourceBalance;
      if ( $Amount >= 0 ) 
      {
        # This is a trade.
          $Withdrawal = $AbsAmount;
      } 
      else 
      {
        # This is a reversal.
          $Deposit = $AbsAmount;
      } 
  } 
  else 
  {
    # Current user is the vendor.
  
      $SetVendorColor = "bgcolor='$LtGreenColor'";
      $Balance = $DestBalance;
      $SetVendorColor = "bgcolor='$LtGreenColor'";
      if ( $Amount >= 0 ) 
      {
        # This is a trade.
          $Deposit = $AbsAmount;
      } 
      else 
      {
        # This is a reversal.
          $Withdrawal = $AbsAmount;
      } 
  };
    
  $self->InsertCommas( '', \$Deposit, \$Withdrawal );
  $self->InsertCommas( 0, \$Balance );
  
  return "
    <tr $SetVendorColor>
      <td align=left> <small>$Date</small> </td>
      <td align=left> <small>$Item</small> </td>
      <td align=left $SetRvsColor> 
        <small>$SourceAcctId</small> </td>
      <td align=left $SetRvsColor> 
        <small>$DestAcctId</small> </td>
      <td align=right $SetRvsColor> 
        <small>$Deposit </small> </td>
      <td align=right $SetRvsColor> 
        <small>$Withdrawal </small> </td>
      <td align=right> 
        <small>$CurrencyId </small> </td>
      <td align=right> 
        <small>$Balance </small> </td>
    </tr>  
  \n";
} 
#-----------------------------------------------
sub InsertCommas    
{ 
  my ( $self, $ShowZeroAs, @NumRefs ) = @_;
  foreach my $NumRef (@NumRefs) 
  {
    my ( $Num, $Segment, @Segments ) = 
      ( abs($$NumRef), 0, () );
    if ( ! $Num ) 
    {
      $$NumRef = $ShowZeroAs;
    } 
    else 
    {
      while ($Num) 
      {
        my $Segment = $Num % 1000;
        $Num = int ( $Num/1000 );
        $Segment = substr( "00$Segment" , -3, 3 ) if $Num;
        unshift @Segments, $Segment;
      };
      $Num = join ',', @Segments;
      $$NumRef = $$NumRef >= 0 ? $Num : "($Num)";
    };
  };
${$NumRefs[0]} }
#-----------------------------------------------
sub InitShowTrades  
{
  my ($self, $acctid, $currencyid) = @_;
  my $vars = $self->{vars};
  my $ccDB            = $self->{ccDB};
  my $ackscols        = $self->{AcksColumns};
  my $n_trades        = $vars->{ShowTradesCount};

  my $limit           = 
    $n_trades eq "All" ? 0 : $n_trades;
  my $latest          = 
    $n_trades eq "All" ? "" : "most recent";
  $vars->{AdminTradeDisplayed}  = 0;
  
  my $ownedcurrencyids  = $vars->{OwnedCurrencyIds};
  my $currencyclause = $currencyid 
    ? qq/AckCurrencyId="$currencyid"/
    : '';
  my $acctidclause = $acctid eq 'TOTAL'
    ? ''
    : qq/(SourceAcctId="$acctid" or DestAcctId="$acctid" )/ ;
    
  my $whereclause = join ' and ', 
    $acctidclause || (), $currencyclause || ();
  
  my $acks = $ccDB->Records( "Acks", 
    -Columns  => $ackscols,
    -Where    => $whereclause,
    -OrderBy  => "AckDateTime desc", 
    -Limit    => $limit
  );
  
  die $self->dump( $vars->{jobwarnings} )
    if $vars->{jobwarnings};
  
  return ( $ownedcurrencyids, $acks, $latest );
  
}
#-----------------------------------------------
sub BoldFieldIfTotal
{
  my ($self, $acctid, $fieldvalue) = @_;
  return $acctid eq 'TOTAL'
    ? "<big><strong>$fieldvalue</strong></big>"
    : $fieldvalue ;
}
#-----------------------------------------------
1;

